AIRFLOW = "airflow"  # Orchestrator
AWS = "aws"  # Key vault / Secrets
AWS_ENDPOINT_S3 = "https://s3." # AWS Artifact Store
AWS_ENDPOINT_URL = ".amazonaws.com" # AWS Artifact Store
AZURE = "azure"  # Artifact store, Key vault
AZUREML = "azureml"  # Step controller (training)
EVIDENTLY = "evidently"  # Profiling
GCP = "gcp"  # Artifact store
KUBEFLOW = "kubeflow"  # Orchestrator, Metadata store
MLFLOW = "mlflow"  # Deployer, logging, context
PYTORCH = "pytorch"  # Special producers
PYTORCH_L = "pytorch_lightning"  # Special producers
S3 = "s3"  # Artifact store
SAGEMAKER = "sagemaker"  # Step controller (training / inference)
SELDON = "seldon"  # Deployer (with Alibi we can do monitoring too!)
SKLEARN = "sklearn"  # Special train steps; Producers
STATSMODELS = "statsmodels" # Producers;
TENSORFLOW = "tensorflow"  # Special train step; Producers
WHYLOGS = "whylogs"  # Data / Model profiling
VERTEX = "vertex"  # Orchestrator, Step controller
XGBOOST = "xgboost" # producers
